""""
Find out what operating system is running.
"""

import platform

from helpers.custom_logger import CustomLogger


class OperatingSystem:
    """
    Find out what operating system is running.
    """

    def __init__(self) -> None:
        self.logger = CustomLogger(self.__class__.__name__)
        self.platform = platform.system()
        self.is_supported_os()

    def __str__(self) -> str:
        """
        Returns the operating system name as a string.
        instead of Object Address
        """
        return self.platform

    def is_supported_os(self):
        """
        Check if the operating system is any OS.
        """
        if not self.platform in ("Linux", "Windows", "Darwin"):
            self.logger.error("Operating system is not supported.")

    def is_mac(self) -> bool:
        """
        Check if the operating system is Mac.
        """
        return self.platform == "Darwin"

    def is_windows(self) -> bool:
        """
        Check if the operating system is Windows.
        """
        return self.platform == "Windows"

    def is_linux(self) -> bool:
        """
        Check if the operating system is Linux.
        """
        return self.platform == "Linux"



"""
Main application file with the main function.
And argument parser.
"""
import asyncio
from argparse import ArgumentParser
from os import path
from shutil import rmtree

from helpers import websocket_server


class Kenobi:
    """
    Class Main which contains the main function.
    And argument parser.
    """

    def __init__(self) -> None:
        """
        Initialize the main class.
        """
        self.debug = False
        self.background = True
        self.logs = False
        self.main()

    @staticmethod
    def parser():
        """
        Parse the arguments.
        Arguments:
            Delete logs: (action delete logs)
                -l, --logs
        """
        parser = ArgumentParser(description="Kenobi Server")
        # add --background and --debug flags default to False
        parser.add_argument("-l", "--logs", action="store_true",
                            default=False, help="Delete logs")
        return parser.parse_args()

    @staticmethod
    def delete_logs():
        """
        Delete logs folder
        """
        print("Deleting logs...")
        # Logs are stored in ~/.kenobi/logs
        # Delete logs folder entirely
        rmtree(path.expanduser("~/.kenobi/logs"))
        print("Done!")

    def main(self):
        """
        Main function.
        """
        args = self.parser()
        self.logs = args.logs

        if self.logs:
            self.delete_logs()
            return
        print("Starting kenobi...")
        websocket_server.WebsocketServer(debug=self.debug)
        asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    Kenobi()



import socket

def get_local_ip():
    ip = "Unknown"
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()
    return ip
    

def get_pairing_code():
    ip = get_local_ip()
    return ip.split(".")[-1]


# pylint: skip-file

import sys
import os
import pytest
from random import randint, choice, shuffle
import string

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from helpers.message_parser import MessageParser

def test_exception():
    """
    Test if exceptions are thrown correctly
    """
    message = "Hello World!"
    parser = MessageParser()
    try:
        _ = parser.parse(message)
    except ValueError:
        assert True
    except:
        assert False

def test_parse_message():
    """
    Test if the message is parsed correctly
    """
    message = "PING:Hello"
    parser = MessageParser()
    parsed = parser.parse(message)
    assert parsed == ("PING", "Hello")

def test_parse_message_with_space():
    """
    Test if the message is parsed correctly
    """
    message = "PING: Hello"
    parser = MessageParser()
    parsed = parser.parse(message)
    assert parsed == ("PING", " Hello")


# Loop through all KeyHeaders and test if they are parsed correctly
headers = [
    'PING',
    "KEY",
    "MOUSE",
    "MEDIA",
    "CLICK",
    "SITE",
    "POWER",
    "LAUNCHAPP",
]    

@pytest.mark.parametrize("test_input, _", zip(headers, headers))
def test_parse_message_with_colon(test_input, _):
    """
    Test if the message is parsed correctly
    """
    message = test_input + ":COMMAND"
    parser = MessageParser()
    parsed = parser.parse(message)
    assert parsed == (test_input, "COMMAND")


mediaCommands = [ "left", "right", "up", "down", "space", "tab", "return", "escape", "playpause", "next", "previous", "mute", "volumeup", "volumedown"]
@pytest.mark.parametrize("test_input, _", zip(mediaCommands, mediaCommands))
def test_parse_message_with_media_command(test_input, _):
    """
    Test if the message is parsed correctly
    """
    message = "MEDIA:" + test_input
    parser = MessageParser()
    parsed = parser.parse(message)
    assert parsed == ("MEDIA", test_input)


# generate array with structure like this -32.4@51.12 24.0@-12.77
def generate_coordinates():
    """
    Generate random coordinates
    """
    x = randint(-1000, 1000)
    y = randint(-1000, 1000)
    return str(x) + "@" + str(y)

coordinates = [generate_coordinates() for _ in range(50)]

@pytest.mark.parametrize("test_input, _", zip(coordinates, coordinates))
def test_parse_message_with_coordinates(test_input, _):
    """
    Test if the message is parsed correctly
    """
    message = test_input
    parser = MessageParser()
    parsed = parser.extract_x_y(message)
    assert parsed == (test_input.split("@")[0], test_input.split("@")[1])


# generate random words 
def generate_words():
    """
    Generate random words
    """
    return ''.join(choice(string.ascii_letters) for _ in range(randint(10, 100)))
    
words = [generate_words() for _ in range(25)]

words = words + headers
shuffle(words)

@pytest.mark.parametrize("test_input, _", zip(words, words))
def test_parse_message_with_words(test_input, _):
    """
    Test if the message is parsed correctly
    """
    message = test_input
    parser = MessageParser()
    if test_input in headers:
        parsed = parser.extract_key(message)
        assert parsed == test_input
    else:
        try:
            parsed = parser.parse(message)
            assert parsed[0] == None or parsed[1] == None
        except ValueError:
            assert True
        except:
            assert False



# pylint: skip-file

from platform import platform
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from helpers.operating_system import OperatingSystem

plat = sys.platform.lower()
osp = OperatingSystem()

def test_os_name():
    if plat == 'linux':
        assert osp.is_linux()
    elif plat == "windows":
        assert osp.is_windows()
    elif plat == 'darwin':
        assert osp.is_mac()

def test_os_version_name():
    if plat == 'linux':
        assert osp.platform == 'Linux'
    elif plat == "windows":
        assert osp.platform == 'Windows'
    elif plat == 'darwin':
        assert osp.platform == 'Darwin'

def test_os_string_modifier():
    if plat == 'Linux':
        assert str(osp) == platform()
    elif plat == "Windows":
        assert str(osp) == platform()
    elif plat == 'Darwin':
        assert str(osp) == platform()




"""
File containing class which represents the emulator.
Controlling media, open apps and links, playing audio, etc is done here.
"""
import webbrowser
from subprocess import run
from os import system

from playsound import playsound
from pynput.keyboard import Controller, Key

from .operating_system import OperatingSystem
from .custom_logger import CustomLogger


class Emulator:
    """
    Class which emulates a key presses, controls media playback
    """

    def __init__(self):
        """
        Initialize the key controller
        And valid key dictionaries
        """
        self.logger = CustomLogger(self.__class__.__name__)
        self.operating_system = OperatingSystem()
        self.keyboard = Controller()

        self.valid_direction_keys = {
            "left": Key.left,
            "right": Key.right,
            "up": Key.up,
            "down": Key.down,
            "space": Key.space,
            "tab": Key.tab,
            "return": Key.enter,
            "escape": Key.esc,
        }

        self.valid_media_keys = {
            "playpause": Key.media_play_pause,
            "next": Key.media_next,
            "previous": Key.media_previous,
            "mute": Key.media_volume_mute,
            "volumeup": Key.media_volume_up,
            "volumedown": Key.media_volume_down
        }

    def emulate_key(self, received_key: str):
        """
        Check if the key is valid, and if so
        Emulate the key using the keyboard controller
        """

        if received_key in self.valid_direction_keys:
            self.keyboard.press(self.valid_direction_keys[received_key])
        else:
            self.logger.info(f"Invalid key {received_key}")

        if received_key in self.valid_media_keys:
            if self.operating_system.platform == "Darwin":
                self.hid_post_aux_key(received_key)
                return
            self.keyboard.press(self.valid_media_keys[received_key])
        else:
            self.logger.info(f"Invalid key {received_key}")

    def hid_post_aux_key(self, key):
        """
        hid post aux key emulation for macOS
        Pynput is not working on macOS so we use this workaround
        """
        import Quartz
        # NSEvent.h
        NSSystemDefined = 14

        # hidsystem/ev_keymap.h
        sound_up_key = 0
        sound_down_key = 1
        play_key = 16
        next_key = 17
        previous_key = 18
        mute_key = 7

        supportedcmds = {
            'playpause': play_key,
            'next': next_key,
            'previous': previous_key,
            'volumeup': sound_up_key,
            'volumedown': sound_down_key,
            'mute': mute_key
        }
        if key in supportedcmds:
            key = supportedcmds[key]
        else:
            self.logger.error(f"Invalid key {key}")
            return

        def do_key(down):
            """
            Handles the key press (keydown or keyup)
            """
            event = Quartz.NSEvent.otherEventWithType_location_modifierFlags_timestamp_windowNumber_context_subtype_data1_data2_(
                NSSystemDefined,  # type
                (0, 0),  # location
                0xa00 if down else 0xb00,  # flags
                0,  # timestamp
                0,  # window
                0,  # ctx
                8,  # subtype
                (key << 16) | ((0xa if down else 0xb) << 8),  # data1
                -1  # data2
            )

            c_event = event.CGEvent()
            Quartz.CGEventPost(0, c_event)
        do_key(True)
        do_key(False)

    def launch_app(self, app: str):
        """
        Launch the app using the system command
        """
        if app.endswith(".com"):
            self.launch_site(url=app)
        else:
            # TEST Launch apps for other OSs
            launch_args = {
                'Linux': ['xdg-open'],
                'Windows': ['start'],
                'Darwin': ['open', '-a']
            }

            try:
                run(launch_args[self.operating_system.platform] + [app])
            except KeyError:
                print(f"Invalid OS \"{self.operating_system}\"")

    @staticmethod
    def ping(value):
        """
        Play the ping sound ping_sound.wav
        """
        if value == "hello":
            # add hello.wav file and play it
            playsound("assets/ping_sound.wav")
        elif value == "ping":
            playsound("assets/ping_sound.wav")
        else:
            print(f"Invalid ping option {value}")

    @staticmethod
    def launch_site(url):
        """
        Launch the site using the system command
        """
        webbrowser.open_new(url=f"https://{url}")

    def power_option(self, value):
        """
        Emulate the power option
        """
        if value == "shutdown":
            self.shutdown()
        elif value == "logout":
            self.logout()
        elif value == "restart":
            self.restart()
        elif value == "sleep":
            self.sleep()
        else:
            print(f"Invalid power option {value}")

    # power options child functions
    def shutdown(self):
        """
        Shutdown the computer depending on the OS
        """
        if self.operating_system.platform == "Linux":
            system("shutdown -h now")
        elif self.operating_system.platform == "Windows":
            system("shutdown -s")
        elif self.operating_system.platform == "Darwin":
            system("shutdown -h now")

    def logout(self):
        """
        Logout of the current session
        """
        if self.operating_system.platform == "Linux":
            system("gnome-session-quit --force")
        elif self.operating_system.platform == "Windows":
            system("shutdown -l")
        elif self.operating_system.platform == "Darwin":
            system("shutdown -l")

    def restart(self):
        """
        Restart the computer
        """
        if self.operating_system.platform == "Linux":
            system("shutdown -r now")
        elif self.operating_system.platform == "Windows":
            system("shutdown -r")
        elif self.operating_system.platform == "Darwin":
            system("shutdown -r now")

    def sleep(self):
        """
        Sleep the computer
        """
        if self.operating_system.platform == "Linux":
            system("systemctl suspend")
        elif self.operating_system.platform == "Windows":
            system("rundll32.exe powrprof.dll,SetSuspendState 0,1,0")
        elif self.operating_system.platform == "Darwin":
            system("pmset sleepnow")



"""
Parses messages from the client and returns the appropriate response
"""
from helpers.custom_logger import CustomLogger

class MessageParser:
    """
    Parse raw messages received from the client
    """

    def __init__(self):
        self.logger = CustomLogger(self.__class__.__name__)
        self.keys = [
            'PING',
            "KEY",
            "MOUSE",
            "MEDIA",
            "CLICK",
            "SITE",
            "POWER",
            "LAUNCHAPP",
        ]

    @staticmethod
    def extract_key(message):
        """
        Returns the key of the message
        Parameters:
            message: the message to be parsed (PING:hello)
        Returns:
            key: the key of the message (PING)
        """
        return message.split(':')[0]

    @staticmethod
    def extract_value(message):
        """
        Returns the value of the message
        Parameters:
            message: the message to be parsed (PING:hello)
        Returns:
            value: the value of the message (hello)
        """
        return message.split(':')[1]

    def validate_key(self, key):
        """
        validates the key by checking if it is in the list of keys
        """
        return key in self.keys

    def parse(self, message):
        """
        Parses the message and returns a tuple of the key and value
        Parameters:
            message: the message to be parsed (PING:hello)
        Returns:
            key: the key of the message (PING)
        """
        if len(message.split(':')) == 2:
            key = self.extract_key(message)
            value = self.extract_value(message)
            if self.validate_key(key):
                return key, value

            self.logger.debug(f"Invalid key: {key}")
            raise ValueError(f"Invalid key: {key}")
        self.logger.debug(f"Invalid message: {message}")
        raise ValueError(f"Invalid data received: {message}")

    def extract_x_y(self, value):
        """
        Returns the x and y coordinates of the mouse
        Parameters:
            value: the value of the message (-13.0@0.0)
        Returns:
            x: the x coordinate (-13.0, 0.0)
        """
        if '@' in value:
            mouse_x, mouse_y = value.split('@')
            return mouse_x, mouse_y
        self.logger.info(f"Invalid data received: {value}")
        raise ValueError(f"Invalid data received: {value}")



"""
File that contains the CustomLogger class.
"""

import logging
from os import path, makedirs


class CustomLogger:
    """
    Custom logger for the application.
    """

    def __init__(self, name, debug=False):
        """
        Initialize the logger
        """
        self.initialize_logger(name, debug)

    def initialize_logger(self, name, debug=False):
        """
        Initialize the logger
        """
        self.logger = logging.getLogger(name)
        self.logger.addHandler(logging.StreamHandler())
        # Save the log file as a .log file in the logs directory
        # Kinda counter-intuitive, will fix later
        if not debug:
            self.logger.setLevel(logging.WARN)
        else:
            self.logger.setLevel(logging.INFO)
        # Store logs in user's home directory in .kenobi/logs
        try:
            self.logger.addHandler(logging.FileHandler(
                f"{path.expanduser('~')}/.kenobi/logs/{name}.log", "a"))
            self.logger.info("Logger file initialized for file %s", name)
        except FileNotFoundError:
            # make the logs directory if it doesn't exist
            try:
                makedirs(f"{path.expanduser('~')}/.kenobi/logs")
                self.logger.addHandler(logging.FileHandler(
                    f"{path.expanduser('~')}/.kenobi/logs/{name}.log", "a"))
                self.logger.info("Logger folder initialized for file %s", name)
            except FileExistsError:
                self.logger.error(
                    "Logger could not be initialized. Logs directory not found.")
            except PermissionError:
                self.logger.error(
                    "Logger could not be initialized. Permission denied.")
        except PermissionError:
            self.logger.error(
                "Logger could not be initialized. Permission denied.")
        except Exception as error:
            self.logger.error(
                "Logger could not be initialized. An unknown error occurred.")
            self.logger.error(error)

    def info(self, message):
        """
        log an info message
        """
        self.logger.info(message)

    def error(self, message):
        """
        log an error message
        """
        self.logger.error(message)

    def warning(self, message):
        """
        log a warning message
        """
        self.logger.warning(message)

    def debug(self, message):
        """
        log a debug message
        """
        self.logger.debug(message)

    def critical(self, message):
        """
        log a critical message
        """
        self.logger.critical(message)



"""
Webserver for the the client to connect to
"""

import asyncio

import websockets

from .custom_logger import CustomLogger
from .emulator import Emulator
from .local_ip import get_local_ip, get_pairing_code
from .message_parser import MessageParser as mp

HOST = "0.0.0.0"
PORT = 9999


class WebsocketServer:
    """
    Class for handling the websocket server and executing the commands
    """

    def __init__(self, debug: bool = False) -> None:
        self.parser = mp()
        self.emulator = Emulator()
        self.logger = CustomLogger(self.__class__.__name__, debug)
        # Server settings (constants)
        # Client Info
        self.connected_ip = None
        self.connected_port = None
        # Start server
        self.logger.warning(f"websocket initialized @ {get_local_ip()}:9999")
        self.logger.warning(f"\nYour pairing code is {get_pairing_code()}\n")
        asyncio.get_event_loop().run_until_complete(self.start())

    async def start(self) -> None:
        """
        Start the websocket server and listen for connections
        """
        self.logger.info(
            f"Starting websocket server @ {HOST}:{PORT}")
        self.server = await websockets.serve(self.handler, HOST, PORT)

    async def handler(self, websocket) -> None:
        """
        Handle the websocket connection, and receive messages
        """
        self.connected_ip = websocket.remote_address[0]
        self.connected_port = websocket.remote_address[1]
        self.logger.warning(
            f"Client connected @ {self.connected_ip}:{self.connected_port}")

        while True:
            try:
                message = await websocket.recv()
                try:
                    key, value = self.parser.parse(message)
                    print(f"Parsed: {key}:{value}")
                    if key == "LAUNCHAPP":
                        self.emulator.launch_app(value)
                    elif key == "PING":
                        self.emulator.ping(value)
                    elif key == "SITE":
                        self.emulator.launch_site(value)
                    elif key == "POWER":
                        self.emulator.power_option(value)
                    elif key in ("MEDIA", "KEY"):
                        self.emulator.emulate_key(value)
                    else:
                        self.logger.info(f"Invalid key: {key}")
                except ValueError as error:
                    self.logger.info(f"Invalid message: {error}")

            except websockets.exceptions.ConnectionClosed:
                self.logger.warning(
                    f"Client disconnected @ {self.connected_ip}:{self.connected_port}")
                self.connected_ip = None
                self.connected_port = None
                break


if __name__ == "__main__":
    server = WebsocketServer()
    asyncio.get_event_loop().run_forever()
